/****************************************************************************
** Meta object code from reading C++ file 'TasEditorWindow.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../fceux/src/drivers/Qt/TasEditor/TasEditorWindow.h"
#include <QtGui/qtextcursor.h>
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'TasEditorWindow.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN20bookmarkPreviewPopupE_t {};
} // unnamed namespace

template <> constexpr inline auto bookmarkPreviewPopup::qt_create_metaobjectdata<qt_meta_tag_ZN20bookmarkPreviewPopupE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "bookmarkPreviewPopup",
        "periodicUpdate",
        "",
        "imageIndexChanged"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'periodicUpdate'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'imageIndexChanged'
        QtMocHelpers::SlotData<void(int)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<bookmarkPreviewPopup, qt_meta_tag_ZN20bookmarkPreviewPopupE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject bookmarkPreviewPopup::staticMetaObject = { {
    QMetaObject::SuperData::link<QDialog::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20bookmarkPreviewPopupE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20bookmarkPreviewPopupE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN20bookmarkPreviewPopupE_t>.metaTypes,
    nullptr
} };

void bookmarkPreviewPopup::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<bookmarkPreviewPopup *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->periodicUpdate(); break;
        case 1: _t->imageIndexChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        default: ;
        }
    }
}

const QMetaObject *bookmarkPreviewPopup::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *bookmarkPreviewPopup::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN20bookmarkPreviewPopupE_t>.strings))
        return static_cast<void*>(this);
    return QDialog::qt_metacast(_clname);
}

int bookmarkPreviewPopup::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDialog::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 2)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 2;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 2)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 2;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN15markerDragPopupE_t {};
} // unnamed namespace

template <> constexpr inline auto markerDragPopup::qt_create_metaobjectdata<qt_meta_tag_ZN15markerDragPopupE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "markerDragPopup",
        "fadeAway",
        ""
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'fadeAway'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPrivate, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<markerDragPopup, qt_meta_tag_ZN15markerDragPopupE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject markerDragPopup::staticMetaObject = { {
    QMetaObject::SuperData::link<QDialog::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15markerDragPopupE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15markerDragPopupE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN15markerDragPopupE_t>.metaTypes,
    nullptr
} };

void markerDragPopup::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<markerDragPopup *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->fadeAway(); break;
        default: ;
        }
    }
    (void)_a;
}

const QMetaObject *markerDragPopup::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *markerDragPopup::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15markerDragPopupE_t>.strings))
        return static_cast<void*>(this);
    return QDialog::qt_metacast(_clname);
}

int markerDragPopup::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDialog::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN10QPianoRollE_t {};
} // unnamed namespace

template <> constexpr inline auto QPianoRoll::qt_create_metaobjectdata<qt_meta_tag_ZN10QPianoRollE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "QPianoRoll",
        "hbarChanged",
        "",
        "val",
        "vbarChanged",
        "setupMarkerDrag"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'hbarChanged'
        QtMocHelpers::SlotData<void(int)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 3 },
        }}),
        // Slot 'vbarChanged'
        QtMocHelpers::SlotData<void(int)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 3 },
        }}),
        // Slot 'setupMarkerDrag'
        QtMocHelpers::SlotData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<QPianoRoll, qt_meta_tag_ZN10QPianoRollE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject QPianoRoll::staticMetaObject = { {
    QMetaObject::SuperData::link<QWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10QPianoRollE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10QPianoRollE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN10QPianoRollE_t>.metaTypes,
    nullptr
} };

void QPianoRoll::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<QPianoRoll *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->hbarChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 1: _t->vbarChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 2: _t->setupMarkerDrag(); break;
        default: ;
        }
    }
}

const QMetaObject *QPianoRoll::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *QPianoRoll::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10QPianoRollE_t>.strings))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int QPianoRoll::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 3)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 3;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 3)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 3;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN18PianoRollScrollBarE_t {};
} // unnamed namespace

template <> constexpr inline auto PianoRollScrollBar::qt_create_metaobjectdata<qt_meta_tag_ZN18PianoRollScrollBarE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "PianoRollScrollBar"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PianoRollScrollBar, qt_meta_tag_ZN18PianoRollScrollBarE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject PianoRollScrollBar::staticMetaObject = { {
    QMetaObject::SuperData::link<QScrollBar::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN18PianoRollScrollBarE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN18PianoRollScrollBarE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN18PianoRollScrollBarE_t>.metaTypes,
    nullptr
} };

void PianoRollScrollBar::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PianoRollScrollBar *>(_o);
    (void)_t;
    (void)_c;
    (void)_id;
    (void)_a;
}

const QMetaObject *PianoRollScrollBar::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *PianoRollScrollBar::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN18PianoRollScrollBarE_t>.strings))
        return static_cast<void*>(this);
    return QScrollBar::qt_metacast(_clname);
}

int PianoRollScrollBar::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QScrollBar::qt_metacall(_c, _id, _a);
    return _id;
}
namespace {
struct qt_meta_tag_ZN22TasRecentProjectActionE_t {};
} // unnamed namespace

template <> constexpr inline auto TasRecentProjectAction::qt_create_metaobjectdata<qt_meta_tag_ZN22TasRecentProjectActionE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "TasRecentProjectAction",
        "activateCB",
        ""
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'activateCB'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TasRecentProjectAction, qt_meta_tag_ZN22TasRecentProjectActionE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject TasRecentProjectAction::staticMetaObject = { {
    QMetaObject::SuperData::link<QAction::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22TasRecentProjectActionE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22TasRecentProjectActionE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN22TasRecentProjectActionE_t>.metaTypes,
    nullptr
} };

void TasRecentProjectAction::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TasRecentProjectAction *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->activateCB(); break;
        default: ;
        }
    }
    (void)_a;
}

const QMetaObject *TasRecentProjectAction::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *TasRecentProjectAction::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22TasRecentProjectActionE_t>.strings))
        return static_cast<void*>(this);
    return QAction::qt_metacast(_clname);
}

int TasRecentProjectAction::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAction::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN17TasFindNoteWindowE_t {};
} // unnamed namespace

template <> constexpr inline auto TasFindNoteWindow::qt_create_metaobjectdata<qt_meta_tag_ZN17TasFindNoteWindowE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "TasFindNoteWindow",
        "closeWindow",
        "",
        "matchCaseChanged",
        "upDirectionSelected",
        "downDirectionSelected",
        "findNextClicked",
        "searchPatternChanged"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'closeWindow'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'matchCaseChanged'
        QtMocHelpers::SlotData<void(bool)>(3, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'upDirectionSelected'
        QtMocHelpers::SlotData<void()>(4, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'downDirectionSelected'
        QtMocHelpers::SlotData<void()>(5, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'findNextClicked'
        QtMocHelpers::SlotData<void()>(6, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'searchPatternChanged'
        QtMocHelpers::SlotData<void(const QString &)>(7, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TasFindNoteWindow, qt_meta_tag_ZN17TasFindNoteWindowE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject TasFindNoteWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<QDialog::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17TasFindNoteWindowE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17TasFindNoteWindowE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN17TasFindNoteWindowE_t>.metaTypes,
    nullptr
} };

void TasFindNoteWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TasFindNoteWindow *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->closeWindow(); break;
        case 1: _t->matchCaseChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 2: _t->upDirectionSelected(); break;
        case 3: _t->downDirectionSelected(); break;
        case 4: _t->findNextClicked(); break;
        case 5: _t->searchPatternChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        default: ;
        }
    }
}

const QMetaObject *TasFindNoteWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *TasFindNoteWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17TasFindNoteWindowE_t>.strings))
        return static_cast<void*>(this);
    return QDialog::qt_metacast(_clname);
}

int TasFindNoteWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDialog::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 6)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 6)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 6;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN17TasEditorSplitterE_t {};
} // unnamed namespace

template <> constexpr inline auto TasEditorSplitter::qt_create_metaobjectdata<qt_meta_tag_ZN17TasEditorSplitterE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "TasEditorSplitter"
    };

    QtMocHelpers::UintData qt_methods {
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TasEditorSplitter, qt_meta_tag_ZN17TasEditorSplitterE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject TasEditorSplitter::staticMetaObject = { {
    QMetaObject::SuperData::link<QSplitter::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17TasEditorSplitterE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17TasEditorSplitterE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN17TasEditorSplitterE_t>.metaTypes,
    nullptr
} };

void TasEditorSplitter::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TasEditorSplitter *>(_o);
    (void)_t;
    (void)_c;
    (void)_id;
    (void)_a;
}

const QMetaObject *TasEditorSplitter::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *TasEditorSplitter::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17TasEditorSplitterE_t>.strings))
        return static_cast<void*>(this);
    return QSplitter::qt_metacast(_clname);
}

int TasEditorSplitter::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QSplitter::qt_metacall(_c, _id, _a);
    return _id;
}
namespace {
struct qt_meta_tag_ZN15TasEditorWindowE_t {};
} // unnamed namespace

template <> constexpr inline auto TasEditorWindow::qt_create_metaobjectdata<qt_meta_tag_ZN15TasEditorWindowE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "TasEditorWindow",
        "closeWindow",
        "",
        "frameUpdate",
        "updateCheckedItems",
        "updateRecordStatus",
        "openProject",
        "saveProjectCb",
        "saveProjectAsCb",
        "saveProjectCompactCb",
        "createNewProject",
        "importMovieFile",
        "exportMovieFile",
        "openOnlineDocs",
        "recordingChanged",
        "recordInputChanged",
        "superImposedChanged",
        "usePatternChanged",
        "playbackPauseCB",
        "playbackFrameRewind",
        "playbackFrameForward",
        "playbackFrameRewindFull",
        "playbackFrameForwardFull",
        "scrollSelectionUpOne",
        "scrollSelectionDnOne",
        "editUndoCB",
        "editRedoCB",
        "editUndoSelCB",
        "editRedoSelCB",
        "editDeselectAll",
        "editSelectAll",
        "editSelBtwMkrs",
        "editReselectClipboard",
        "editCutCB",
        "editCopyCB",
        "editPasteCB",
        "editPasteInsertCB",
        "editClearCB",
        "editDeleteCB",
        "editCloneCB",
        "editInsertCB",
        "editInsertNumFramesCB",
        "editTruncateMovieCB",
        "openFindNoteWindow",
        "dpyBrnchScrnChanged",
        "dpyBrnchDescChanged",
        "enaHotChgChanged",
        "followUndoActChanged",
        "followMkrActChanged",
        "enaGrnznActChanged",
        "afPtrnSkipLagActChanged",
        "adjInputLagActChanged",
        "drawInputDragActChanged",
        "cmbRecDrawActChanged",
        "use1PforRecActChanged",
        "useInputColSetActChanged",
        "bindMkrInputActChanged",
        "emptyNewMkrNotesActChanged",
        "oldCtlBrnhSchemeActChanged",
        "brnchRestoreMovieActChanged",
        "hudInScrnBranchActChanged",
        "pauseAtEndActChanged",
        "showToolTipsActChanged",
        "upperMarkerLabelClicked",
        "lowerMarkerLabelClicked",
        "histTreeItemActivated",
        "QTreeWidgetItem*",
        "playbackFollowCursorCb",
        "playbackAutoRestoreCb",
        "playbackTurboSeekCb",
        "openProjectSaveOptions",
        "setGreenzoneCapacity",
        "setMaxUndoCapacity",
        "setCurrentPattern",
        "tabViewChanged",
        "findSimilarNote",
        "findNextSimilarNote",
        "jumpToPreviousMarker",
        "jumpToNextMarker",
        "openAboutWindow",
        "autoLuaRunChanged",
        "manLuaRun",
        "setMarkers",
        "removeMarkers",
        "ungreenzoneSelectedFrames",
        "activateHotkey",
        "hkIdx",
        "QShortcut*",
        "shortcut",
        "changePianoRollFontCB",
        "changeBookmarksFontCB",
        "changeBranchesFontCB"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'closeWindow'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'frameUpdate'
        QtMocHelpers::SlotData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'updateCheckedItems'
        QtMocHelpers::SlotData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'updateRecordStatus'
        QtMocHelpers::SlotData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'openProject'
        QtMocHelpers::SlotData<void()>(6, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveProjectCb'
        QtMocHelpers::SlotData<void()>(7, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveProjectAsCb'
        QtMocHelpers::SlotData<void()>(8, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveProjectCompactCb'
        QtMocHelpers::SlotData<void()>(9, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'createNewProject'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'importMovieFile'
        QtMocHelpers::SlotData<void()>(11, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'exportMovieFile'
        QtMocHelpers::SlotData<void()>(12, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openOnlineDocs'
        QtMocHelpers::SlotData<void()>(13, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'recordingChanged'
        QtMocHelpers::SlotData<void(int)>(14, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'recordInputChanged'
        QtMocHelpers::SlotData<void(int)>(15, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'superImposedChanged'
        QtMocHelpers::SlotData<void(int)>(16, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'usePatternChanged'
        QtMocHelpers::SlotData<void(int)>(17, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'playbackPauseCB'
        QtMocHelpers::SlotData<void()>(18, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'playbackFrameRewind'
        QtMocHelpers::SlotData<void()>(19, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'playbackFrameForward'
        QtMocHelpers::SlotData<void()>(20, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'playbackFrameRewindFull'
        QtMocHelpers::SlotData<void()>(21, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'playbackFrameForwardFull'
        QtMocHelpers::SlotData<void()>(22, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'scrollSelectionUpOne'
        QtMocHelpers::SlotData<void()>(23, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'scrollSelectionDnOne'
        QtMocHelpers::SlotData<void()>(24, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editUndoCB'
        QtMocHelpers::SlotData<void()>(25, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editRedoCB'
        QtMocHelpers::SlotData<void()>(26, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editUndoSelCB'
        QtMocHelpers::SlotData<void()>(27, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editRedoSelCB'
        QtMocHelpers::SlotData<void()>(28, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editDeselectAll'
        QtMocHelpers::SlotData<void()>(29, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editSelectAll'
        QtMocHelpers::SlotData<void()>(30, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editSelBtwMkrs'
        QtMocHelpers::SlotData<void()>(31, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editReselectClipboard'
        QtMocHelpers::SlotData<void()>(32, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editCutCB'
        QtMocHelpers::SlotData<void()>(33, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editCopyCB'
        QtMocHelpers::SlotData<void()>(34, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editPasteCB'
        QtMocHelpers::SlotData<void()>(35, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editPasteInsertCB'
        QtMocHelpers::SlotData<void()>(36, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editClearCB'
        QtMocHelpers::SlotData<void()>(37, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editDeleteCB'
        QtMocHelpers::SlotData<void()>(38, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editCloneCB'
        QtMocHelpers::SlotData<void()>(39, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editInsertCB'
        QtMocHelpers::SlotData<void()>(40, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editInsertNumFramesCB'
        QtMocHelpers::SlotData<void()>(41, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'editTruncateMovieCB'
        QtMocHelpers::SlotData<void()>(42, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openFindNoteWindow'
        QtMocHelpers::SlotData<void()>(43, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'dpyBrnchScrnChanged'
        QtMocHelpers::SlotData<void(bool)>(44, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'dpyBrnchDescChanged'
        QtMocHelpers::SlotData<void(bool)>(45, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'enaHotChgChanged'
        QtMocHelpers::SlotData<void(bool)>(46, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'followUndoActChanged'
        QtMocHelpers::SlotData<void(bool)>(47, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'followMkrActChanged'
        QtMocHelpers::SlotData<void(bool)>(48, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'enaGrnznActChanged'
        QtMocHelpers::SlotData<void(bool)>(49, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'afPtrnSkipLagActChanged'
        QtMocHelpers::SlotData<void(bool)>(50, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'adjInputLagActChanged'
        QtMocHelpers::SlotData<void(bool)>(51, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'drawInputDragActChanged'
        QtMocHelpers::SlotData<void(bool)>(52, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'cmbRecDrawActChanged'
        QtMocHelpers::SlotData<void(bool)>(53, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'use1PforRecActChanged'
        QtMocHelpers::SlotData<void(bool)>(54, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'useInputColSetActChanged'
        QtMocHelpers::SlotData<void(bool)>(55, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'bindMkrInputActChanged'
        QtMocHelpers::SlotData<void(bool)>(56, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'emptyNewMkrNotesActChanged'
        QtMocHelpers::SlotData<void(bool)>(57, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'oldCtlBrnhSchemeActChanged'
        QtMocHelpers::SlotData<void(bool)>(58, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'brnchRestoreMovieActChanged'
        QtMocHelpers::SlotData<void(bool)>(59, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'hudInScrnBranchActChanged'
        QtMocHelpers::SlotData<void(bool)>(60, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'pauseAtEndActChanged'
        QtMocHelpers::SlotData<void(bool)>(61, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'showToolTipsActChanged'
        QtMocHelpers::SlotData<void(bool)>(62, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'upperMarkerLabelClicked'
        QtMocHelpers::SlotData<void()>(63, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'lowerMarkerLabelClicked'
        QtMocHelpers::SlotData<void()>(64, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'histTreeItemActivated'
        QtMocHelpers::SlotData<void(QTreeWidgetItem *, int)>(65, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 66, 2 }, { QMetaType::Int, 2 },
        }}),
        // Slot 'playbackFollowCursorCb'
        QtMocHelpers::SlotData<void(bool)>(67, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'playbackAutoRestoreCb'
        QtMocHelpers::SlotData<void(bool)>(68, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'playbackTurboSeekCb'
        QtMocHelpers::SlotData<void(bool)>(69, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'openProjectSaveOptions'
        QtMocHelpers::SlotData<void()>(70, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setGreenzoneCapacity'
        QtMocHelpers::SlotData<void()>(71, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setMaxUndoCapacity'
        QtMocHelpers::SlotData<void()>(72, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setCurrentPattern'
        QtMocHelpers::SlotData<void(int)>(73, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'tabViewChanged'
        QtMocHelpers::SlotData<void(int)>(74, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'findSimilarNote'
        QtMocHelpers::SlotData<void()>(75, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'findNextSimilarNote'
        QtMocHelpers::SlotData<void()>(76, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'jumpToPreviousMarker'
        QtMocHelpers::SlotData<void()>(77, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'jumpToNextMarker'
        QtMocHelpers::SlotData<void()>(78, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openAboutWindow'
        QtMocHelpers::SlotData<void()>(79, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'autoLuaRunChanged'
        QtMocHelpers::SlotData<void(bool)>(80, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'manLuaRun'
        QtMocHelpers::SlotData<void()>(81, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setMarkers'
        QtMocHelpers::SlotData<void()>(82, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'removeMarkers'
        QtMocHelpers::SlotData<void()>(83, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'ungreenzoneSelectedFrames'
        QtMocHelpers::SlotData<void()>(84, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'activateHotkey'
        QtMocHelpers::SlotData<void(int, QShortcut *)>(85, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 86 }, { 0x80000000 | 87, 88 },
        }}),
        // Slot 'changePianoRollFontCB'
        QtMocHelpers::SlotData<void()>(89, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeBookmarksFontCB'
        QtMocHelpers::SlotData<void()>(90, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeBranchesFontCB'
        QtMocHelpers::SlotData<void()>(91, 2, QMC::AccessPrivate, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TasEditorWindow, qt_meta_tag_ZN15TasEditorWindowE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject TasEditorWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<QDialog::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15TasEditorWindowE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15TasEditorWindowE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN15TasEditorWindowE_t>.metaTypes,
    nullptr
} };

void TasEditorWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TasEditorWindow *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->closeWindow(); break;
        case 1: _t->frameUpdate(); break;
        case 2: _t->updateCheckedItems(); break;
        case 3: _t->updateRecordStatus(); break;
        case 4: _t->openProject(); break;
        case 5: _t->saveProjectCb(); break;
        case 6: _t->saveProjectAsCb(); break;
        case 7: _t->saveProjectCompactCb(); break;
        case 8: _t->createNewProject(); break;
        case 9: _t->importMovieFile(); break;
        case 10: _t->exportMovieFile(); break;
        case 11: _t->openOnlineDocs(); break;
        case 12: _t->recordingChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 13: _t->recordInputChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 14: _t->superImposedChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 15: _t->usePatternChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 16: _t->playbackPauseCB(); break;
        case 17: _t->playbackFrameRewind(); break;
        case 18: _t->playbackFrameForward(); break;
        case 19: _t->playbackFrameRewindFull(); break;
        case 20: _t->playbackFrameForwardFull(); break;
        case 21: _t->scrollSelectionUpOne(); break;
        case 22: _t->scrollSelectionDnOne(); break;
        case 23: _t->editUndoCB(); break;
        case 24: _t->editRedoCB(); break;
        case 25: _t->editUndoSelCB(); break;
        case 26: _t->editRedoSelCB(); break;
        case 27: _t->editDeselectAll(); break;
        case 28: _t->editSelectAll(); break;
        case 29: _t->editSelBtwMkrs(); break;
        case 30: _t->editReselectClipboard(); break;
        case 31: _t->editCutCB(); break;
        case 32: _t->editCopyCB(); break;
        case 33: _t->editPasteCB(); break;
        case 34: _t->editPasteInsertCB(); break;
        case 35: _t->editClearCB(); break;
        case 36: _t->editDeleteCB(); break;
        case 37: _t->editCloneCB(); break;
        case 38: _t->editInsertCB(); break;
        case 39: _t->editInsertNumFramesCB(); break;
        case 40: _t->editTruncateMovieCB(); break;
        case 41: _t->openFindNoteWindow(); break;
        case 42: _t->dpyBrnchScrnChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 43: _t->dpyBrnchDescChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 44: _t->enaHotChgChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 45: _t->followUndoActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 46: _t->followMkrActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 47: _t->enaGrnznActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 48: _t->afPtrnSkipLagActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 49: _t->adjInputLagActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 50: _t->drawInputDragActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 51: _t->cmbRecDrawActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 52: _t->use1PforRecActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 53: _t->useInputColSetActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 54: _t->bindMkrInputActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 55: _t->emptyNewMkrNotesActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 56: _t->oldCtlBrnhSchemeActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 57: _t->brnchRestoreMovieActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 58: _t->hudInScrnBranchActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 59: _t->pauseAtEndActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 60: _t->showToolTipsActChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 61: _t->upperMarkerLabelClicked(); break;
        case 62: _t->lowerMarkerLabelClicked(); break;
        case 63: _t->histTreeItemActivated((*reinterpret_cast<std::add_pointer_t<QTreeWidgetItem*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 64: _t->playbackFollowCursorCb((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 65: _t->playbackAutoRestoreCb((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 66: _t->playbackTurboSeekCb((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 67: _t->openProjectSaveOptions(); break;
        case 68: _t->setGreenzoneCapacity(); break;
        case 69: _t->setMaxUndoCapacity(); break;
        case 70: _t->setCurrentPattern((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 71: _t->tabViewChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 72: _t->findSimilarNote(); break;
        case 73: _t->findNextSimilarNote(); break;
        case 74: _t->jumpToPreviousMarker(); break;
        case 75: _t->jumpToNextMarker(); break;
        case 76: _t->openAboutWindow(); break;
        case 77: _t->autoLuaRunChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 78: _t->manLuaRun(); break;
        case 79: _t->setMarkers(); break;
        case 80: _t->removeMarkers(); break;
        case 81: _t->ungreenzoneSelectedFrames(); break;
        case 82: _t->activateHotkey((*reinterpret_cast<std::add_pointer_t<int>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QShortcut*>>(_a[2]))); break;
        case 83: _t->changePianoRollFontCB(); break;
        case 84: _t->changeBookmarksFontCB(); break;
        case 85: _t->changeBranchesFontCB(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 82:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 1:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QShortcut* >(); break;
            }
            break;
        }
    }
}

const QMetaObject *TasEditorWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *TasEditorWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN15TasEditorWindowE_t>.strings))
        return static_cast<void*>(this);
    return QDialog::qt_metacast(_clname);
}

int TasEditorWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDialog::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 86)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 86;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 86)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 86;
    }
    return _id;
}
QT_WARNING_POP
