// fceux_git_info.cpp -- DO NOT EDIT: This file is auto-generated at build
#include "Qt/fceux_git_info.h" 
#define FCEUX_GIT_URL  "https://github.com/TASEmulators/fc" 
#define FCEUX_GIT_REV  "v2.6.6" 
const char *fceu_get_git_url(void){ return FCEUX_GIT_URL; }
const char *fceu_get_git_rev(void){ return FCEUX_GIT_REV; }
