/****************************************************************************
** Meta object code from reading C++ file 'ConsoleWindow.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../fceux/src/drivers/Qt/ConsoleWindow.h"
#include <QtGui/qtextcursor.h>
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'ConsoleWindow.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN16emulatorThread_tE_t {};
} // unnamed namespace

template <> constexpr inline auto emulatorThread_t::qt_create_metaobjectdata<qt_meta_tag_ZN16emulatorThread_tE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "emulatorThread_t",
        "finished",
        "",
        "frameFinished",
        "loadRomRequest",
        "s"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'finished'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'frameFinished'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'loadRomRequest'
        QtMocHelpers::SignalData<void(QString)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 5 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<emulatorThread_t, qt_meta_tag_ZN16emulatorThread_tE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject emulatorThread_t::staticMetaObject = { {
    QMetaObject::SuperData::link<QThread::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN16emulatorThread_tE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN16emulatorThread_tE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN16emulatorThread_tE_t>.metaTypes,
    nullptr
} };

void emulatorThread_t::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<emulatorThread_t *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->finished(); break;
        case 1: _t->frameFinished(); break;
        case 2: _t->loadRomRequest((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (emulatorThread_t::*)()>(_a, &emulatorThread_t::finished, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (emulatorThread_t::*)()>(_a, &emulatorThread_t::frameFinished, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (emulatorThread_t::*)(QString )>(_a, &emulatorThread_t::loadRomRequest, 2))
            return;
    }
}

const QMetaObject *emulatorThread_t::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *emulatorThread_t::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN16emulatorThread_tE_t>.strings))
        return static_cast<void*>(this);
    return QThread::qt_metacast(_clname);
}

int emulatorThread_t::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QThread::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 3)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 3;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 3)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 3;
    }
    return _id;
}

// SIGNAL 0
void emulatorThread_t::finished()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void emulatorThread_t::frameFinished()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void emulatorThread_t::loadRomRequest(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}
namespace {
struct qt_meta_tag_ZN18autoFireMenuActionE_t {};
} // unnamed namespace

template <> constexpr inline auto autoFireMenuAction::qt_create_metaobjectdata<qt_meta_tag_ZN18autoFireMenuActionE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "autoFireMenuAction",
        "activateCB",
        ""
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'activateCB'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<autoFireMenuAction, qt_meta_tag_ZN18autoFireMenuActionE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject autoFireMenuAction::staticMetaObject = { {
    QMetaObject::SuperData::link<QAction::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN18autoFireMenuActionE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN18autoFireMenuActionE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN18autoFireMenuActionE_t>.metaTypes,
    nullptr
} };

void autoFireMenuAction::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<autoFireMenuAction *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->activateCB(); break;
        default: ;
        }
    }
    (void)_a;
}

const QMetaObject *autoFireMenuAction::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *autoFireMenuAction::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN18autoFireMenuActionE_t>.strings))
        return static_cast<void*>(this);
    return QAction::qt_metacast(_clname);
}

int autoFireMenuAction::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAction::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN22consoleRecentRomActionE_t {};
} // unnamed namespace

template <> constexpr inline auto consoleRecentRomAction::qt_create_metaobjectdata<qt_meta_tag_ZN22consoleRecentRomActionE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "consoleRecentRomAction",
        "activateCB",
        ""
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'activateCB'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<consoleRecentRomAction, qt_meta_tag_ZN22consoleRecentRomActionE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject consoleRecentRomAction::staticMetaObject = { {
    QMetaObject::SuperData::link<QAction::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22consoleRecentRomActionE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22consoleRecentRomActionE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN22consoleRecentRomActionE_t>.metaTypes,
    nullptr
} };

void consoleRecentRomAction::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<consoleRecentRomAction *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->activateCB(); break;
        default: ;
        }
    }
    (void)_a;
}

const QMetaObject *consoleRecentRomAction::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *consoleRecentRomAction::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN22consoleRecentRomActionE_t>.strings))
        return static_cast<void*>(this);
    return QAction::qt_metacast(_clname);
}

int consoleRecentRomAction::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAction::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN12consoleWin_tE_t {};
} // unnamed namespace

template <> constexpr inline auto consoleWin_t::qt_create_metaobjectdata<qt_meta_tag_ZN12consoleWin_tE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "consoleWin_t",
        "openDebugWindow",
        "",
        "openHexEditor",
        "openGamePadConfWin",
        "toggleFullscreen",
        "toggleMenuVis",
        "recordMovie",
        "winResizeIx",
        "iScale",
        "closeApp",
        "openROMFile",
        "loadNSF",
        "loadStateFrom",
        "saveStateAs",
        "quickLoad",
        "quickSave",
        "closeROMCB",
        "aboutFCEUX",
        "aboutQt",
        "openOnlineDocs",
        "openOfflineDocs",
        "openTasEditor",
        "openMsgLogWin",
        "openInputConfWin",
        "openGameSndConfWin",
        "openGameVideoConfWin",
        "openHotkeyConfWin",
        "openPaletteConfWin",
        "openGuiConfWin",
        "openTimingConfWin",
        "openStateRecorderConfWin",
        "openPaletteEditorWin",
        "openAviRiffViewer",
        "openTimingStatWin",
        "openMovieOptWin",
        "openCodeDataLogger",
        "openTraceLogger",
        "openFamilyKeyboard",
        "toggleAutoResume",
        "updatePeriodic",
        "changeState0",
        "changeState1",
        "changeState2",
        "changeState3",
        "changeState4",
        "changeState5",
        "changeState6",
        "changeState7",
        "changeState8",
        "changeState9",
        "incrementState",
        "decrementState",
        "loadLua",
        "takeScreenShot",
        "prepareScreenShot",
        "powerConsoleCB",
        "consoleHardReset",
        "consoleSoftReset",
        "consolePause",
        "toggleGameGenie",
        "checked",
        "loadGameGenieROM",
        "loadMostRecentROM",
        "setRegionNTSC",
        "setRegionPAL",
        "setRegionDendy",
        "setRamInit0",
        "setRamInit1",
        "setRamInit2",
        "setRamInit3",
        "insertCoin",
        "fdsSwitchDisk",
        "fdsEjectDisk",
        "fdsLoadBiosFile",
        "emuSpeedUp",
        "emuSlowDown",
        "emuSlowestSpd",
        "emuNormalSpd",
        "emuFastestSpd",
        "emuCustomSpd",
        "emuSetFrameAdvDelay",
        "openPPUViewer",
        "openOAMViewer",
        "openNTViewer",
        "openGGEncoder",
        "openNesHeaderEditor",
        "openCheats",
        "openRamWatch",
        "openRamSearch",
        "openMovie",
        "stopMovie",
        "playMovieFromBeginning",
        "setCustomAutoFire",
        "muteSoundVolume",
        "incrSoundVolume",
        "decrSoundVolume",
        "toggleLagCounterDisplay",
        "toggleFrameAdvLagSkip",
        "toggleMovieBindSaveState",
        "toggleMovieFrameDisplay",
        "toggleMovieReadWrite",
        "toggleInputDisplay",
        "toggleTurboMode",
        "toggleBackground",
        "toggleForeground",
        "toggleFamKeyBrdEnable",
        "toggleGlobalCheatEnable",
        "saveState0",
        "saveState1",
        "saveState2",
        "saveState3",
        "saveState4",
        "saveState5",
        "saveState6",
        "saveState7",
        "saveState8",
        "saveState9",
        "loadState0",
        "loadState1",
        "loadState2",
        "loadState3",
        "loadState4",
        "loadState5",
        "loadState6",
        "loadState7",
        "loadState8",
        "loadState9",
        "loadPrevState",
        "loadNextState",
        "mainMenuOpen",
        "mainMenuClose",
        "warnAmbiguousShortcut",
        "QShortcut*",
        "aviRecordStart",
        "aviRecordAsStart",
        "aviRecordStop",
        "aviAudioEnableChange",
        "aviVideoFormatChanged",
        "idx",
        "setAviHudEnable",
        "setAviMsgEnable",
        "wavRecordStart",
        "wavRecordAsStart",
        "wavRecordStop",
        "winScreenChanged",
        "QScreen*",
        "scr",
        "winActiveChanged",
        "emuFrameFinish",
        "toggleMenuAutoHide",
        "toggleUseBgPaletteForVideo",
        "videoBgColorChanged",
        "QColor&",
        "c",
        "loadRomRequestCB",
        "s",
        "videoDriverDestroyed",
        "obj"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'openDebugWindow'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'openHexEditor'
        QtMocHelpers::SlotData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'openGamePadConfWin'
        QtMocHelpers::SlotData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'toggleFullscreen'
        QtMocHelpers::SlotData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'toggleMenuVis'
        QtMocHelpers::SlotData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'recordMovie'
        QtMocHelpers::SlotData<void()>(7, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'winResizeIx'
        QtMocHelpers::SlotData<void(int)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 9 },
        }}),
        // Slot 'closeApp'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openROMFile'
        QtMocHelpers::SlotData<void()>(11, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadNSF'
        QtMocHelpers::SlotData<void()>(12, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadStateFrom'
        QtMocHelpers::SlotData<void()>(13, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveStateAs'
        QtMocHelpers::SlotData<void()>(14, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'quickLoad'
        QtMocHelpers::SlotData<void()>(15, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'quickSave'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'closeROMCB'
        QtMocHelpers::SlotData<void()>(17, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'aboutFCEUX'
        QtMocHelpers::SlotData<void()>(18, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'aboutQt'
        QtMocHelpers::SlotData<void()>(19, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openOnlineDocs'
        QtMocHelpers::SlotData<void()>(20, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openOfflineDocs'
        QtMocHelpers::SlotData<void()>(21, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openTasEditor'
        QtMocHelpers::SlotData<void()>(22, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openMsgLogWin'
        QtMocHelpers::SlotData<void()>(23, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openInputConfWin'
        QtMocHelpers::SlotData<void()>(24, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openGameSndConfWin'
        QtMocHelpers::SlotData<void()>(25, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openGameVideoConfWin'
        QtMocHelpers::SlotData<void()>(26, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openHotkeyConfWin'
        QtMocHelpers::SlotData<void()>(27, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openPaletteConfWin'
        QtMocHelpers::SlotData<void()>(28, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openGuiConfWin'
        QtMocHelpers::SlotData<void()>(29, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openTimingConfWin'
        QtMocHelpers::SlotData<void()>(30, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openStateRecorderConfWin'
        QtMocHelpers::SlotData<void()>(31, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openPaletteEditorWin'
        QtMocHelpers::SlotData<void()>(32, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openAviRiffViewer'
        QtMocHelpers::SlotData<void()>(33, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openTimingStatWin'
        QtMocHelpers::SlotData<void()>(34, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openMovieOptWin'
        QtMocHelpers::SlotData<void()>(35, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openCodeDataLogger'
        QtMocHelpers::SlotData<void()>(36, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openTraceLogger'
        QtMocHelpers::SlotData<void()>(37, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openFamilyKeyboard'
        QtMocHelpers::SlotData<void()>(38, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleAutoResume'
        QtMocHelpers::SlotData<void()>(39, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'updatePeriodic'
        QtMocHelpers::SlotData<void()>(40, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState0'
        QtMocHelpers::SlotData<void()>(41, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState1'
        QtMocHelpers::SlotData<void()>(42, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState2'
        QtMocHelpers::SlotData<void()>(43, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState3'
        QtMocHelpers::SlotData<void()>(44, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState4'
        QtMocHelpers::SlotData<void()>(45, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState5'
        QtMocHelpers::SlotData<void()>(46, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState6'
        QtMocHelpers::SlotData<void()>(47, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState7'
        QtMocHelpers::SlotData<void()>(48, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState8'
        QtMocHelpers::SlotData<void()>(49, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'changeState9'
        QtMocHelpers::SlotData<void()>(50, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'incrementState'
        QtMocHelpers::SlotData<void()>(51, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'decrementState'
        QtMocHelpers::SlotData<void()>(52, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadLua'
        QtMocHelpers::SlotData<void()>(53, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'takeScreenShot'
        QtMocHelpers::SlotData<void()>(54, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'prepareScreenShot'
        QtMocHelpers::SlotData<void()>(55, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'powerConsoleCB'
        QtMocHelpers::SlotData<void()>(56, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'consoleHardReset'
        QtMocHelpers::SlotData<void()>(57, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'consoleSoftReset'
        QtMocHelpers::SlotData<void()>(58, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'consolePause'
        QtMocHelpers::SlotData<void()>(59, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleGameGenie'
        QtMocHelpers::SlotData<void(bool)>(60, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 61 },
        }}),
        // Slot 'loadGameGenieROM'
        QtMocHelpers::SlotData<void()>(62, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadMostRecentROM'
        QtMocHelpers::SlotData<void()>(63, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setRegionNTSC'
        QtMocHelpers::SlotData<void()>(64, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setRegionPAL'
        QtMocHelpers::SlotData<void()>(65, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setRegionDendy'
        QtMocHelpers::SlotData<void()>(66, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setRamInit0'
        QtMocHelpers::SlotData<void()>(67, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setRamInit1'
        QtMocHelpers::SlotData<void()>(68, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setRamInit2'
        QtMocHelpers::SlotData<void()>(69, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setRamInit3'
        QtMocHelpers::SlotData<void()>(70, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'insertCoin'
        QtMocHelpers::SlotData<void()>(71, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'fdsSwitchDisk'
        QtMocHelpers::SlotData<void()>(72, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'fdsEjectDisk'
        QtMocHelpers::SlotData<void()>(73, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'fdsLoadBiosFile'
        QtMocHelpers::SlotData<void()>(74, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'emuSpeedUp'
        QtMocHelpers::SlotData<void()>(75, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'emuSlowDown'
        QtMocHelpers::SlotData<void()>(76, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'emuSlowestSpd'
        QtMocHelpers::SlotData<void()>(77, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'emuNormalSpd'
        QtMocHelpers::SlotData<void()>(78, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'emuFastestSpd'
        QtMocHelpers::SlotData<void()>(79, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'emuCustomSpd'
        QtMocHelpers::SlotData<void()>(80, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'emuSetFrameAdvDelay'
        QtMocHelpers::SlotData<void()>(81, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openPPUViewer'
        QtMocHelpers::SlotData<void()>(82, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openOAMViewer'
        QtMocHelpers::SlotData<void()>(83, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openNTViewer'
        QtMocHelpers::SlotData<void()>(84, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openGGEncoder'
        QtMocHelpers::SlotData<void()>(85, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openNesHeaderEditor'
        QtMocHelpers::SlotData<void()>(86, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openCheats'
        QtMocHelpers::SlotData<void()>(87, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openRamWatch'
        QtMocHelpers::SlotData<void()>(88, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openRamSearch'
        QtMocHelpers::SlotData<void()>(89, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'openMovie'
        QtMocHelpers::SlotData<void()>(90, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'stopMovie'
        QtMocHelpers::SlotData<void()>(91, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'playMovieFromBeginning'
        QtMocHelpers::SlotData<void()>(92, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'setCustomAutoFire'
        QtMocHelpers::SlotData<void()>(93, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'muteSoundVolume'
        QtMocHelpers::SlotData<void()>(94, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'incrSoundVolume'
        QtMocHelpers::SlotData<void()>(95, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'decrSoundVolume'
        QtMocHelpers::SlotData<void()>(96, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleLagCounterDisplay'
        QtMocHelpers::SlotData<void()>(97, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleFrameAdvLagSkip'
        QtMocHelpers::SlotData<void()>(98, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleMovieBindSaveState'
        QtMocHelpers::SlotData<void()>(99, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleMovieFrameDisplay'
        QtMocHelpers::SlotData<void()>(100, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleMovieReadWrite'
        QtMocHelpers::SlotData<void()>(101, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleInputDisplay'
        QtMocHelpers::SlotData<void()>(102, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleTurboMode'
        QtMocHelpers::SlotData<void()>(103, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleBackground'
        QtMocHelpers::SlotData<void()>(104, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleForeground'
        QtMocHelpers::SlotData<void()>(105, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleFamKeyBrdEnable'
        QtMocHelpers::SlotData<void()>(106, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleGlobalCheatEnable'
        QtMocHelpers::SlotData<void()>(107, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState0'
        QtMocHelpers::SlotData<void()>(108, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState1'
        QtMocHelpers::SlotData<void()>(109, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState2'
        QtMocHelpers::SlotData<void()>(110, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState3'
        QtMocHelpers::SlotData<void()>(111, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState4'
        QtMocHelpers::SlotData<void()>(112, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState5'
        QtMocHelpers::SlotData<void()>(113, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState6'
        QtMocHelpers::SlotData<void()>(114, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState7'
        QtMocHelpers::SlotData<void()>(115, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState8'
        QtMocHelpers::SlotData<void()>(116, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'saveState9'
        QtMocHelpers::SlotData<void()>(117, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState0'
        QtMocHelpers::SlotData<void()>(118, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState1'
        QtMocHelpers::SlotData<void()>(119, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState2'
        QtMocHelpers::SlotData<void()>(120, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState3'
        QtMocHelpers::SlotData<void()>(121, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState4'
        QtMocHelpers::SlotData<void()>(122, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState5'
        QtMocHelpers::SlotData<void()>(123, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState6'
        QtMocHelpers::SlotData<void()>(124, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState7'
        QtMocHelpers::SlotData<void()>(125, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState8'
        QtMocHelpers::SlotData<void()>(126, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadState9'
        QtMocHelpers::SlotData<void()>(127, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadPrevState'
        QtMocHelpers::SlotData<void()>(128, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'loadNextState'
        QtMocHelpers::SlotData<void()>(129, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'mainMenuOpen'
        QtMocHelpers::SlotData<void()>(130, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'mainMenuClose'
        QtMocHelpers::SlotData<void()>(131, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'warnAmbiguousShortcut'
        QtMocHelpers::SlotData<void(QShortcut *)>(132, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 133, 2 },
        }}),
        // Slot 'aviRecordStart'
        QtMocHelpers::SlotData<void()>(134, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'aviRecordAsStart'
        QtMocHelpers::SlotData<void()>(135, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'aviRecordStop'
        QtMocHelpers::SlotData<void()>(136, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'aviAudioEnableChange'
        QtMocHelpers::SlotData<void(bool)>(137, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'aviVideoFormatChanged'
        QtMocHelpers::SlotData<void(int)>(138, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Int, 139 },
        }}),
        // Slot 'setAviHudEnable'
        QtMocHelpers::SlotData<void(bool)>(140, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setAviMsgEnable'
        QtMocHelpers::SlotData<void(bool)>(141, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'wavRecordStart'
        QtMocHelpers::SlotData<void()>(142, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'wavRecordAsStart'
        QtMocHelpers::SlotData<void()>(143, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'wavRecordStop'
        QtMocHelpers::SlotData<void()>(144, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'winScreenChanged'
        QtMocHelpers::SlotData<void(QScreen *)>(145, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 146, 147 },
        }}),
        // Slot 'winActiveChanged'
        QtMocHelpers::SlotData<void()>(148, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'emuFrameFinish'
        QtMocHelpers::SlotData<void()>(149, 2, QMC::AccessPrivate, QMetaType::Void),
        // Slot 'toggleMenuAutoHide'
        QtMocHelpers::SlotData<void(bool)>(150, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'toggleUseBgPaletteForVideo'
        QtMocHelpers::SlotData<void(bool)>(151, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'videoBgColorChanged'
        QtMocHelpers::SlotData<void(QColor &)>(152, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { 0x80000000 | 153, 154 },
        }}),
        // Slot 'loadRomRequestCB'
        QtMocHelpers::SlotData<void(QString)>(155, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QString, 156 },
        }}),
        // Slot 'videoDriverDestroyed'
        QtMocHelpers::SlotData<void(QObject *)>(157, 2, QMC::AccessPrivate, QMetaType::Void, {{
            { QMetaType::QObjectStar, 158 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<consoleWin_t, qt_meta_tag_ZN12consoleWin_tE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject consoleWin_t::staticMetaObject = { {
    QMetaObject::SuperData::link<QMainWindow::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12consoleWin_tE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12consoleWin_tE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN12consoleWin_tE_t>.metaTypes,
    nullptr
} };

void consoleWin_t::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<consoleWin_t *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->openDebugWindow(); break;
        case 1: _t->openHexEditor(); break;
        case 2: _t->openGamePadConfWin(); break;
        case 3: _t->toggleFullscreen(); break;
        case 4: _t->toggleMenuVis(); break;
        case 5: _t->recordMovie(); break;
        case 6: _t->winResizeIx((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 7: _t->closeApp(); break;
        case 8: _t->openROMFile(); break;
        case 9: _t->loadNSF(); break;
        case 10: _t->loadStateFrom(); break;
        case 11: _t->saveStateAs(); break;
        case 12: _t->quickLoad(); break;
        case 13: _t->quickSave(); break;
        case 14: _t->closeROMCB(); break;
        case 15: _t->aboutFCEUX(); break;
        case 16: _t->aboutQt(); break;
        case 17: _t->openOnlineDocs(); break;
        case 18: _t->openOfflineDocs(); break;
        case 19: _t->openTasEditor(); break;
        case 20: _t->openMsgLogWin(); break;
        case 21: _t->openInputConfWin(); break;
        case 22: _t->openGameSndConfWin(); break;
        case 23: _t->openGameVideoConfWin(); break;
        case 24: _t->openHotkeyConfWin(); break;
        case 25: _t->openPaletteConfWin(); break;
        case 26: _t->openGuiConfWin(); break;
        case 27: _t->openTimingConfWin(); break;
        case 28: _t->openStateRecorderConfWin(); break;
        case 29: _t->openPaletteEditorWin(); break;
        case 30: _t->openAviRiffViewer(); break;
        case 31: _t->openTimingStatWin(); break;
        case 32: _t->openMovieOptWin(); break;
        case 33: _t->openCodeDataLogger(); break;
        case 34: _t->openTraceLogger(); break;
        case 35: _t->openFamilyKeyboard(); break;
        case 36: _t->toggleAutoResume(); break;
        case 37: _t->updatePeriodic(); break;
        case 38: _t->changeState0(); break;
        case 39: _t->changeState1(); break;
        case 40: _t->changeState2(); break;
        case 41: _t->changeState3(); break;
        case 42: _t->changeState4(); break;
        case 43: _t->changeState5(); break;
        case 44: _t->changeState6(); break;
        case 45: _t->changeState7(); break;
        case 46: _t->changeState8(); break;
        case 47: _t->changeState9(); break;
        case 48: _t->incrementState(); break;
        case 49: _t->decrementState(); break;
        case 50: _t->loadLua(); break;
        case 51: _t->takeScreenShot(); break;
        case 52: _t->prepareScreenShot(); break;
        case 53: _t->powerConsoleCB(); break;
        case 54: _t->consoleHardReset(); break;
        case 55: _t->consoleSoftReset(); break;
        case 56: _t->consolePause(); break;
        case 57: _t->toggleGameGenie((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 58: _t->loadGameGenieROM(); break;
        case 59: _t->loadMostRecentROM(); break;
        case 60: _t->setRegionNTSC(); break;
        case 61: _t->setRegionPAL(); break;
        case 62: _t->setRegionDendy(); break;
        case 63: _t->setRamInit0(); break;
        case 64: _t->setRamInit1(); break;
        case 65: _t->setRamInit2(); break;
        case 66: _t->setRamInit3(); break;
        case 67: _t->insertCoin(); break;
        case 68: _t->fdsSwitchDisk(); break;
        case 69: _t->fdsEjectDisk(); break;
        case 70: _t->fdsLoadBiosFile(); break;
        case 71: _t->emuSpeedUp(); break;
        case 72: _t->emuSlowDown(); break;
        case 73: _t->emuSlowestSpd(); break;
        case 74: _t->emuNormalSpd(); break;
        case 75: _t->emuFastestSpd(); break;
        case 76: _t->emuCustomSpd(); break;
        case 77: _t->emuSetFrameAdvDelay(); break;
        case 78: _t->openPPUViewer(); break;
        case 79: _t->openOAMViewer(); break;
        case 80: _t->openNTViewer(); break;
        case 81: _t->openGGEncoder(); break;
        case 82: _t->openNesHeaderEditor(); break;
        case 83: _t->openCheats(); break;
        case 84: _t->openRamWatch(); break;
        case 85: _t->openRamSearch(); break;
        case 86: _t->openMovie(); break;
        case 87: _t->stopMovie(); break;
        case 88: _t->playMovieFromBeginning(); break;
        case 89: _t->setCustomAutoFire(); break;
        case 90: _t->muteSoundVolume(); break;
        case 91: _t->incrSoundVolume(); break;
        case 92: _t->decrSoundVolume(); break;
        case 93: _t->toggleLagCounterDisplay(); break;
        case 94: _t->toggleFrameAdvLagSkip(); break;
        case 95: _t->toggleMovieBindSaveState(); break;
        case 96: _t->toggleMovieFrameDisplay(); break;
        case 97: _t->toggleMovieReadWrite(); break;
        case 98: _t->toggleInputDisplay(); break;
        case 99: _t->toggleTurboMode(); break;
        case 100: _t->toggleBackground(); break;
        case 101: _t->toggleForeground(); break;
        case 102: _t->toggleFamKeyBrdEnable(); break;
        case 103: _t->toggleGlobalCheatEnable(); break;
        case 104: _t->saveState0(); break;
        case 105: _t->saveState1(); break;
        case 106: _t->saveState2(); break;
        case 107: _t->saveState3(); break;
        case 108: _t->saveState4(); break;
        case 109: _t->saveState5(); break;
        case 110: _t->saveState6(); break;
        case 111: _t->saveState7(); break;
        case 112: _t->saveState8(); break;
        case 113: _t->saveState9(); break;
        case 114: _t->loadState0(); break;
        case 115: _t->loadState1(); break;
        case 116: _t->loadState2(); break;
        case 117: _t->loadState3(); break;
        case 118: _t->loadState4(); break;
        case 119: _t->loadState5(); break;
        case 120: _t->loadState6(); break;
        case 121: _t->loadState7(); break;
        case 122: _t->loadState8(); break;
        case 123: _t->loadState9(); break;
        case 124: _t->loadPrevState(); break;
        case 125: _t->loadNextState(); break;
        case 126: _t->mainMenuOpen(); break;
        case 127: _t->mainMenuClose(); break;
        case 128: _t->warnAmbiguousShortcut((*reinterpret_cast<std::add_pointer_t<QShortcut*>>(_a[1]))); break;
        case 129: _t->aviRecordStart(); break;
        case 130: _t->aviRecordAsStart(); break;
        case 131: _t->aviRecordStop(); break;
        case 132: _t->aviAudioEnableChange((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 133: _t->aviVideoFormatChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 134: _t->setAviHudEnable((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 135: _t->setAviMsgEnable((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 136: _t->wavRecordStart(); break;
        case 137: _t->wavRecordAsStart(); break;
        case 138: _t->wavRecordStop(); break;
        case 139: _t->winScreenChanged((*reinterpret_cast<std::add_pointer_t<QScreen*>>(_a[1]))); break;
        case 140: _t->winActiveChanged(); break;
        case 141: _t->emuFrameFinish(); break;
        case 142: _t->toggleMenuAutoHide((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 143: _t->toggleUseBgPaletteForVideo((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 144: _t->videoBgColorChanged((*reinterpret_cast<std::add_pointer_t<QColor&>>(_a[1]))); break;
        case 145: _t->loadRomRequestCB((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 146: _t->videoDriverDestroyed((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 128:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QShortcut* >(); break;
            }
            break;
        case 139:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QScreen* >(); break;
            }
            break;
        }
    }
}

const QMetaObject *consoleWin_t::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *consoleWin_t::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12consoleWin_tE_t>.strings))
        return static_cast<void*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int consoleWin_t::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 147)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 147;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 147)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 147;
    }
    return _id;
}
QT_WARNING_POP
