#ifndef EMUFILE_TYPES_H
#define EMUFILE_TYPES_H

#include "types.h"

typedef uint8 u8;
typedef uint16 u16;
typedef uint32 u32;
typedef uint64 u64;
typedef int8 s8;
typedef int16 s16;
typedef int32 s32;
typedef int64 s64;


#endif //EMUFILE_TYPES_H
